/*
 * @Author: kael 
 * @Date: 2020-11-05 18:19:03 
 * @Last Modified by: kael
 * @Last Modified time: 2021-11-09 10:43:03
 */

const logger = require('../logger')();
const Fetch = require('../fetch');
const Api = require('../api');

// https://github.com/Marak/colors.js/blob/56de9f0983f68cd0a08c5b76d10a783e4b881716/lib/styles.js
const codes = {
  black: [30, 39],
  red: [31, 39],
  green: [32, 39],
  yellow: [33, 39],
  blue: [34, 39],
  magenta: [35, 39],
  cyan: [36, 39],
  white: [37, 39],
  gray: [90, 39],
};

function colors(string, name) {
  let [open, close] = codes[name];
  return `\u001b[${open}m${string}\u001b[${close}m`;
}

function moving_average(datas, days = 3) {
  if (!Array.isArray(datas)) return 0;
  let total = datas
    .slice(0, days)
    .map(v => parseFloat(v[2]))
    .reduce((sum, v) => sum += v, 0);
  return total / days;
}

// MA3 = MA5
// => (tm2 + tm1 + tm0) / 3 = (tm4 + tm3 + tm2 + tm1 + tm0) / 5
// => (tm2 + tm1 + tm0) * 5 = (tm4 + tm3 + tm2 + tm1 + tm0) * 3
// => 5tm2 + 5tm1 + 5tm0 - 3tm2 - 3tm1 - 3tm0 = 3tm4 + 3tm3
// => 2tm2 + 2tm1 + 2tm0 = 3tm4 + 3tm3
// => 2tm0 = 3tm4 + 3tm3 - 2tm2 - 2tm1
// => tm0 = 1.5tm4 + 1.5tm3 - tm2 - tm1
function marginal_value(datas) {
  let [tm4, tm3, tm2, tm1, tm0] = datas.slice(0, 5).map(v => parseFloat(v[2])).reverse();
  return {
    now: tm0,
    curr: 1.5 * tm4 + 1.5 * tm3 - tm2 - tm1,
    next: 1.5 * tm3 + 1.5 * tm2 - tm1 - tm0,
  };
}

const padding = [
  16,
  20,
  15,
  15,
  15,
  15,
  15,
  { size: 15, color: (v) => /^\+/.test(v) ? 'cyan' : 'white' },
  { size: 15, color: (v) => /-\]/.test(v) ? 'blue' : 'white' },
];
const pad = (items) => {
  return items.map((item, index) => {
    let conf = padding[index];
    let size = conf.size || conf;
    let text = item[index ? 'padStart' : 'padEnd'](size - item.split(/[^\x00-\xff]/).length - 1);
    return conf.color ? colors(text, conf.color(item)) : text;
  }).join('');
};

const LINE = new Array(124).fill('=').join('');

const STOCKS_MAP = new Map();
function holding(code) {
  let stock = STOCKS_MAP.get(code);
  return (stock && stock.market_val) ? '*' : '';
}

function roc(MV) {
  let { now, next } = MV;
  let v = 100 * (next - now) / now;
  return (v > 0 ? '+' : '') + v.toFixed(2) + '%';
}

function ceil(v, d = 2) {
  return Math.ceil(Math.abs(v)).toString().padStart(d, '0');
}

function sale_signal(code, now) {
  let stock = STOCKS_MAP.get(code);
  let { cost_price } = stock;
  if (!cost_price) return '[=====][00]';
  let irate = 100 * (now - cost_price) / cost_price;
  let srate = [-25, -15, -8, -5, -2].map(signal => irate < signal ? '-' : ' ').join('');
  return `[${srate}][${ceil(irate)}]`;
}

function start(codes) {
  let quotes = codes.map(code => Api.quote.kline_dayqfq(code, 10));
  Promise.all(quotes).then(resps => {
    console.reset();
    console.log((new Date()).toTimeString().split(' ')[0]);
    console.log(LINE);
    console.log(pad(['ID', '名称', 'NOW', 'MA3', 'MA5', 'MVC', 'MVN', 'MVN-ROC', 'S-SIG']));
    console.log(LINE);
    let rows = [[], LINE.replace(/=/g, '-'), []];
    let errors = resps.filter((data, index) => {
      let { fid, id, info, data: resp } = data;
      if (!id || !info || !resp) return false;
      let name = String(info[1] || '').replace(/(ETF|\s).+$/i, '');
      let MA3 = moving_average(resp, 3);
      let MA5 = moving_average(resp, 5);
      let MV = marginal_value(resp);
      if (!MV.now) return global.MARKET[id] = `us${info[2]}`;
      let hold_star = holding(fid);
      let row = pad([
        fid + hold_star, String(name).replace(/-\w+$/, ''),
        MV.now.toFixed(3), MA3.toFixed(3), MA5.toFixed(3),
        MV.curr.toFixed(3), MV.next.toFixed(3), roc(MV),
        sale_signal(fid, MV.now),
      ]);
      rows[hold_star ? 0 : 2].push(colors(row, MA3 > MA5 ? 'red' : MA3 < MA5 ? 'green' : 'white'))
    });
    console.log(rows.flat().join('\n'));
    console.log(LINE);
    if (errors.length) throw new Error('Errors occur.');
    setTimeout(() => start(codes), 10 * 1000);
  }).catch((error) => {
    logger.error(error.message);
    setTimeout(() => start(codes), 1 * 1000);
  });
}

module.exports = () => {
  Fetch.futu().then(([holders, follows]) => {
    let stocks = [
      ...Object.values(holders),
      ...Object.values(follows),
    ].flat();
    stocks.forEach(stock => STOCKS_MAP.has(stock.code) || STOCKS_MAP.set(stock.code, stock));
    start([...STOCKS_MAP.keys()]);
  });
};